#include "window.h"

#include <libgnome/libgnome.h>

#include "i18n.h"

#include <gdk/gdkx.h>
#include <gdk/gdk.h>

#include <X11/Xatom.h>
#include <X11/Xlib.h>

GNOME_CLASS_BOILERPLATE(DdWindow, dd_window,
                        GtkWindow, GTK_TYPE_WINDOW)

enum {
	DD_WINDOW_SIGNAL_WORKAREA_CHANGED = 0,
	DD_WINDOW_SIGNAL_MAX
};

static guint dd_window_signals[DD_WINDOW_SIGNAL_MAX] = { 0 };

static void dd_window_set_workarea_from_net_workarea(DdWindow *window) {
	GdkScreen    *screen         = gtk_window_get_screen(GTK_WINDOW(window));
	GdkWindow    *root_window    = gdk_screen_get_root_window(screen);
	gint          screen_size[2] = { gdk_screen_get_width(screen), gdk_screen_get_height(screen) };
	GdkAtom       net_workarea   = gdk_atom_intern("_NET_WORKAREA", FALSE);
	GdkAtom       xa_cardinal    = gdk_x11_xatom_to_atom(XA_CARDINAL);
	GdkAtom       type;
	gint          format, length;
	gint32       *workareas      = NULL;
	GdkRectangle  workarea       = window->workarea;

	/* Used Nautilus for this code */
	gdk_error_trap_push();
	if(!gdk_property_get(root_window, net_workarea, xa_cardinal, 0, G_MAXLONG, FALSE,
	                     &type, &format, &length, (guchar **) &workareas))
		workareas = NULL;

	workarea.x = workarea.y = 0;
	workarea.width = screen_size[0];
	workarea.height = screen_size[1];

	if(!gdk_error_trap_pop() && workareas != NULL && type == xa_cardinal && (length % sizeof(workareas[0])) == 0 && format == 32) {
		unsigned int a;

		for(a = 0; a < length/sizeof(workareas[0]); a += 4) {
			gulong x = workareas[a + 0];
			gulong y = workareas[a + 1];
			gulong w = workareas[a + 2];
			gulong h = workareas[a + 3];

			if((x + w) > screen_size[0] || (y + h) > screen_size[1])
				continue;

			if(x > workarea.x) {
				workarea.width += (workarea.x - x);
				workarea.x = x;
			}
			if(y > workarea.y) {
				workarea.height += (workarea.y - y);
				workarea.y = y;
			}
			if((x + w) < (workarea.x + workarea.width))
				workarea.width = w;
			if((y + h) < (workarea.y + workarea.height))
				workarea.height = h;
		}
	}

	if(window->workarea.x != workarea.x || window->workarea.y != workarea.y
	|| window->workarea.width != workarea.width || window->workarea.height != workarea.height) {
		window->workarea = workarea;

		g_signal_emit(window, dd_window_signals[DD_WINDOW_SIGNAL_WORKAREA_CHANGED], 0);
	}

	g_free(workareas);
}

static gint dd_window_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data) {
	return TRUE;
}
static GdkFilterReturn dd_window_root_window_event_filter(GdkXEvent *gdk_xevent, GdkEvent *event, gpointer user_data) {
	DdWindow *window = user_data;
	XEvent   *xevent = gdk_xevent;

	switch(xevent->type) {
		case PropertyNotify: {
			if(xevent->xproperty.atom == gdk_x11_get_xatom_by_name("_NET_WORKAREA"))
				dd_window_set_workarea_from_net_workarea(window);
		}	break;
	}

	return GDK_FILTER_CONTINUE;
}

static void dd_window_instance_init(DdWindow *window) {
	window->old_desktop = 0;

	{ /* Set the original work area to the whole desktop */
		GdkScreen *screen = gtk_window_get_screen(GTK_WINDOW(window));

		window->workarea.x = window->workarea.y = 0;
		window->workarea.width  = gdk_screen_get_width(screen);
		window->workarea.height = gdk_screen_get_height(screen);
	}

	g_signal_connect(window, "delete-event", G_CALLBACK(&dd_window_delete_event), NULL);

	gtk_window_set_title(GTK_WINDOW(window), _("Desktop Dealie"));

	/* Use this window as the desktop */
	gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DESKTOP);

	/* Use this window in each workspace */
	gtk_window_stick(GTK_WINDOW(window));

#if 0
	/* Don't double buffer the window, it's probably transparent anyway */
	if(GTK_WIDGET_DOUBLE_BUFFERED(GTK_WIDGET(window)))
		gtk_widget_set_double_buffered(GTK_WIDGET(window), FALSE);
#endif

	/* Realize the widget before the rest of the program can get at it not-really-initialized */
	if(!GTK_WIDGET_REALIZED(GTK_WIDGET(window)))
		gtk_widget_realize(GTK_WIDGET(window));
}

static void dd_window_finalize(GObject *object) {
	DdWindow *window = DD_WINDOW(object);

	/* ... */

	G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void dd_window_destroy(GtkObject *object) {
	DdWindow *window = DD_WINDOW(object);

	if(GTK_WIDGET_REALIZED(GTK_WIDGET(object)))
		gtk_widget_unrealize(GTK_WIDGET(object));

	GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

/* Stolen from Nautilus: */
static guint get_max_forced_height(GdkScreen *screen) {
	return (gdk_screen_get_height(screen) * 90) / 100;
}
static guint get_max_forced_width(GdkScreen *screen) {
	return (gdk_screen_get_width(screen) * 90) / 100;
}
static void dd_window_size_request(GtkWidget *widget, GtkRequisition *req) {
	GdkScreen *screen;
	guint max_width;
	guint max_height;

	g_assert(DD_IS_WINDOW(widget));
	g_assert(req != NULL);

	GTK_WIDGET_CLASS(parent_class)->size_request(widget, req);

	screen = gtk_window_get_screen(GTK_WINDOW(widget));

	/* Limit the requisition to be within 90% of the available screen 
	 * real state.
	 *
	 * This way the user will have a fighting chance of getting
	 * control of their window back if for whatever reason one of the
	 * window's descendants decide they want to be 4000 pixels wide.
	 *
	 * Note that the user can still make the window really huge by hand.
	 *
	 * Bugs in components or other widgets that cause such huge geometries
	 * to be requested, should still be fixed.  This code is here only to 
	 * prevent the extremely frustrating consequence of such bugs.
	 */
	max_width  = get_max_forced_width(screen);
	max_height = get_max_forced_height(screen);

	if(req->width > (int)max_width)
		req->width = max_width;
	if(req->height > (int)max_height)
		req->height = max_height;
}

static void dd_window_screen_size_changed(GdkScreen *screen, gpointer user_data) {
	DdWindow *window = user_data;

	gtk_widget_set_size_request(GTK_WIDGET(window), gdk_screen_get_width(screen), gdk_screen_get_height(screen));
}

static void dd_window_realize(GtkWidget *widget) {
	DdWindow *window = DD_WINDOW(widget);

	GTK_WIDGET_CLASS(parent_class)->realize(widget);

	{ /* Hack to make other programs happy */
		GdkAtom  nautilus_desktop_window_id = gdk_atom_intern("NAUTILUS_DESKTOP_WINDOW_ID", FALSE);
		GdkAtom  atom;
		gint     format, length;
		guchar  *data = NULL;
		Window   xid;
		
		/* Check if something else that is or is pretending to be Nautilus is running and back up its xid property */
		if(gdk_property_get(gdk_get_default_root_window(), nautilus_desktop_window_id,
			gdk_x11_xatom_to_atom(XA_WINDOW), 0, 4, FALSE, &atom, &format, &length, (guchar **) &data)) {
			xid = *((Window *)data);

			window->old_desktop = xid;

			g_free(data);
		} else {
			window->old_desktop = 0;
		}

		/* Set our own dealie */
		xid = GDK_WINDOW_XID(GTK_WIDGET(window)->window);

		gdk_property_change(gdk_get_default_root_window(), nautilus_desktop_window_id,
			gdk_x11_xatom_to_atom(XA_WINDOW), 32,
			GDK_PROP_MODE_REPLACE, (guchar *) &xid, 1);
	}

	g_signal_connect(gtk_window_get_screen(GTK_WINDOW(window)), "size-changed", G_CALLBACK(&dd_window_screen_size_changed), window);

	{
		GdkScreen *screen      = gtk_window_get_screen(GTK_WINDOW(window));
		GdkWindow *root_window = gdk_screen_get_root_window(screen);

		dd_window_set_workarea_from_net_workarea(window);

		gdk_window_set_events(root_window, GDK_PROPERTY_NOTIFY);
		gdk_window_add_filter(root_window, &dd_window_root_window_event_filter, window);
	}
}

static void dd_window_unrealize(GtkWidget *widget) {
	DdWindow *window = DD_WINDOW(widget);

	{
		GdkScreen *screen      = gtk_window_get_screen(GTK_WINDOW(window));
		GdkWindow *root_window = gdk_screen_get_root_window(screen);

		gdk_window_remove_filter(root_window, &dd_window_root_window_event_filter, window);
	}

	if(window->old_desktop) {
		GdkAtom nautilus_desktop_window_id = gdk_atom_intern("NAUTILUS_DESKTOP_WINDOW_ID", TRUE);

		gdk_property_change(gdk_get_default_root_window(), nautilus_desktop_window_id,
			gdk_x11_xatom_to_atom(XA_WINDOW), 32,
			GDK_PROP_MODE_REPLACE, (guchar *) &window->old_desktop, 1);

		window->old_desktop = 0;
	} else {
		GdkAtom nautilus_desktop_window_id = gdk_atom_intern("NAUTILUS_DESKTOP_WINDOW_ID", TRUE);

		gdk_property_delete(gdk_get_default_root_window(), nautilus_desktop_window_id);
	}

	g_signal_handlers_disconnect_by_func(gtk_window_get_screen(GTK_WINDOW(window)), &dd_window_screen_size_changed, window);
		
	GTK_WIDGET_CLASS(parent_class)->unrealize(widget);
}

static void dd_window_class_init(DdWindowClass *class) {
	G_OBJECT_CLASS(class)->finalize       = &dd_window_finalize;
	GTK_OBJECT_CLASS(class)->destroy      = &dd_window_destroy;
	GTK_WIDGET_CLASS(class)->size_request = &dd_window_size_request;
	GTK_WIDGET_CLASS(class)->realize      = &dd_window_realize;
	GTK_WIDGET_CLASS(class)->unrealize    = &dd_window_unrealize;

	dd_window_signals[DD_WINDOW_SIGNAL_WORKAREA_CHANGED] = g_signal_new("workarea-changed",
		dd_window_get_type(),
		G_SIGNAL_RUN_FIRST,
		G_STRUCT_OFFSET(DdWindowClass, workarea_changed),
		NULL, NULL, 
		&g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
}

DdWindow *dd_window_new(void) {
	GdkScreen *screen = gdk_screen_get_default();
	DdWindow  *window;

	window = DD_WINDOW(gtk_widget_new(dd_window_get_type(), 
		"width_request",  gdk_screen_get_width(screen),
		"height_request", gdk_screen_get_height(screen),
		NULL));

	return window;
}

G_CONST_RETURN GdkRectangle *dd_window_get_workarea(const DdWindow *window) {
	return &window->workarea;
}

void dd_window_add(DdWindow *window, GtkWidget *widget) {
	gtk_container_add(GTK_CONTAINER(window), widget);
}

static void dd_window_fixed_workarea_changed(DdWindow *window, gpointer user_data) {
	GtkFixed  *fixed       = user_data;
	GtkWidget *in_workarea = g_object_get_data(G_OBJECT(fixed), "dd_window_in_workarea_with_fixed");

	if(in_workarea) {
		const GdkRectangle *workarea = dd_window_get_workarea(window);

		gtk_fixed_move(fixed, in_workarea, workarea->x, workarea->y);
		gtk_widget_set_size_request(in_workarea, workarea->width, workarea->height);
	}
}
static void dd_window_fixed_destroy(GtkObject *object, gpointer user_data) {
	GtkFixed *fixed  = GTK_FIXED(object);
	DdWindow *window = user_data;

	g_signal_handlers_disconnect_by_func(window, &dd_window_fixed_workarea_changed, fixed);
}

void dd_window_add_in_workarea_with_fixed(DdWindow *window, GtkWidget *widget) {
	const GdkRectangle *workarea = dd_window_get_workarea(window);
	GtkWidget          *fixed    = gtk_fixed_new();

	g_signal_connect(fixed, "destroy", G_CALLBACK(&dd_window_fixed_destroy), window);
	g_signal_connect(window, "workarea-changed", G_CALLBACK(&dd_window_fixed_workarea_changed), GTK_FIXED(fixed));

	gtk_widget_set_size_request(widget, workarea->width, workarea->height);
	gtk_fixed_put(GTK_FIXED(fixed), widget, workarea->x, workarea->y);

	g_object_set_data(G_OBJECT(fixed), "dd_window_in_workarea_with_fixed", widget);

	dd_window_add(window, fixed);
	gtk_widget_show(fixed);
}
